theory AbsoluteValue
imports Main HOL.Real
begin

definition abs :: "real \<Rightarrow> real" where
  "abs x = (if x \<ge> 0 then x else - x)"

lemma abs_nonneg:
  shows "abs a \<ge> 0"
  unfolding abs_def
  by auto

lemma
  shows "abs a = 0 \<longleftrightarrow> a = 0"
  unfolding abs_def
  by auto

lemma 
  shows "abs (a * b) = abs a * abs b"
  unfolding abs_def
  thm zero_le_mult_iff
  by (simp add: zero_le_mult_iff)

lemma
  shows "abs (a + b) \<le> abs a + abs b"
proof-
  obtain s where "abs (a + b) = s * (a + b)"
                 "s = 1 \<or> s = -1"
    unfolding abs_def
    by (metis mult_1 mult_minus1)

  have "\<forall> x. -1 * x \<le> abs x \<and> 1 * x \<le> abs x" 
    unfolding abs_def
    by auto
  
  then have *: "\<forall> x. s * x \<le> abs x"
    using \<open>s = 1 \<or> s = -1\<close>
    by auto

  have "abs (a + b) = s * (a + b)"
    by fact
  also have "... = s * a + s * b"
    by (simp add: distrib_left)
  also have "... \<le> abs a + abs b"
    using *
    by (smt (verit))
  finally show ?thesis
    .
qed

end
