theory GeneralRecursion
imports Main HOL.Real "HOL-Library.Code_Target_Nat" "HOL-Library.Multiset"
begin

fun (sequential) power :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
   "power x 0 = 1"
 | "power x n = 
       (if n mod 2 = 0 then 
          power (x * x) (n div 2)
        else
          x * power x (n - 1))"

value "power 2 64"

lemma "power x n = x^n"
proof (induction x n rule: power.induct)
  case (1 x)
  show ?case
    by simp
next
  case (2 x n)
  show ?case
  proof (cases "Suc n mod 2 = 0")
    case True
    then have "power x (Suc n) = power (x*x) (Suc n div 2)"
      by simp
    also have "... = (x*x)^(Suc n div 2)"
      using 2 True
      by simp
    finally show ?thesis
      using True
      by (metis dvd_mult_div_cancel mod_0_imp_dvd power2_eq_square power_mult)
  next
    case False
    then have "power x (Suc n) = x * power x n"
      by simp
    also have "... = x * x ^ n"
      using 2 False
      by simp
    finally show ?thesis
      by simp
  qed
qed


fun merge :: "nat list \<Rightarrow> nat list \<Rightarrow> nat list" where
  "merge xs [] = xs"
| "merge [] ys = ys"
| "merge (x # xs) (y # ys) = 
     (if x \<le> y then x # merge xs (y # ys) 
      else y # merge (x # xs) ys)"

lemma mset_merge [simp]:
  shows "mset (merge xs ys) = mset xs + mset ys"
  by (induction xs ys rule: merge.induct) auto

lemma set_merge [simp]:
  shows "set (merge xs ys) = set xs \<union> set ys"
  using mset_merge
  by (metis set_mset_mset set_mset_union)

lemma sorted_merge [simp]:
  assumes "sorted xs" "sorted ys"
  shows "sorted (merge xs ys)"
  using assms
  by (induction xs ys rule: merge.induct) auto
end