theory Isar
imports Main
begin

thm equalityI
thm subsetI

lemma "- (A \<union> B) = -A \<inter> -B"
proof
  show "- (A \<union> B) \<subseteq> -A \<inter> -B"
  proof
    fix x
    assume "x \<in> - (A \<union> B)"
    then have "x \<notin> A \<union> B"
      by (rule ComplD)
    show "x \<in> -A \<inter> -B"
    proof
      show "x \<in> - A"
      proof
        assume "x \<in> A"
        then have "x \<in> A \<union> B"
          by (rule UnI1)
        then show False
          using \<open>x \<notin> A \<union> B\<close>
          by - (erule notE)
      qed
    next
      show "x \<in> - B"
      proof
        assume "x \<in> B"
        then have "x \<in> A \<union> B"
          by (rule UnI2)
        then show False
          using \<open>x \<notin> A \<union> B\<close>
          by - (erule notE)
      qed
    qed
  qed
next
  show "-A \<inter> -B \<subseteq> - (A \<union> B)"
  proof
    fix x
    assume "x \<in> -A \<inter> -B"

    from \<open>x \<in> -A \<inter> -B\<close> have "x \<in> -A"
      by - (erule IntE)                   
    then have "x \<notin> A"
      by (rule ComplD)

    from \<open>x \<in> -A \<inter> -B\<close> have "x \<in> -B"
      by - (erule IntE)
    then have "x \<notin> B"
      by (rule ComplD)

    show "x \<in> - (A \<union> B)"
    proof
      assume "x \<in> A \<union> B"
      then show False
      proof
        assume "x \<in> A" 
        with \<open>x \<notin> A\<close> show False
          by - (erule notE)
      next
        assume "x \<in> B"
        with \<open>x \<notin> B\<close> show False
          by - (erule notE)
      qed
    qed
  qed
qed

lemma "A \<and> B \<longrightarrow> B \<and> A"
proof
  assume "A \<and> B"
  from \<open>A \<and> B\<close> have "A" 
    by - (erule conjE)
  from \<open>A \<and> B\<close> have "B" 
    by - (erule conjE)
  from \<open>A\<close> \<open>B\<close> show "B \<and> A"
    by - (rule conjI)
qed


lemma
  assumes "\<forall> x. \<exists> y. R x y"
  assumes sym: "\<forall> x y. R x y \<longrightarrow> R y x"
  assumes trans: "\<forall> x y z. R x y \<and> R y z \<longrightarrow> R x z"
  shows "\<forall> x. R x x"
proof
  fix x

  obtain y where "R x y"
    using assms(1)          
    by auto

  with sym have "R y x" 
    by auto

  from `R x y` `R y x` 
  show "R x x"
    using trans
    by blast
qed

lemma "\<exists> d. (drinks d \<longrightarrow> (\<forall> x. drinks x))"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then have "\<forall> d. \<not> (drinks d \<longrightarrow> (\<forall> x. drinks x))"
    by auto
  then have "\<forall> d. drinks d \<and> \<not> (\<forall> x. drinks x)"
    by auto
  then have "(\<forall> d. drinks d) \<and> \<not> (\<forall> x. drinks x)"
    by auto
  then show False
    by auto
qed

lemma "\<exists> d. (drinks d \<longrightarrow> (\<forall> x. drinks x))"
proof (cases "\<forall> x. drinks x")
  case True
  then show ?thesis
    by auto
next
  case False
  then obtain d where "\<not> drinks d"
    by auto
  then show ?thesis
    by auto
qed

lemma
  assumes ell: "\<forall> x. (\<exists> y. loves x y) \<longrightarrow> (\<forall> z. loves z x)"
  assumes "loves John Marry"
  shows "loves Iago Othello"
proof-
  have "\<exists> y. loves John y"
    using `loves John Marry`
    by auto
  then have "\<forall> z. loves z John"
    using ell
    by auto
  then have "loves Othello John"
    by auto
  then have "\<exists> y. loves Othello y"
    by auto
  then have "\<forall> z. loves z Othello"
    using ell
    by auto
  then show "loves Iago Othello"
    by auto
qed

lemma
  assumes "inj f" "inj g"
  shows "inj (f \<circ> g)"
  unfolding inj_def
proof safe
  fix x y
  assume "(f \<circ> g) x = (f \<circ> g) y"
  then have "f (g x) = f (g y)"
    unfolding comp_def
    by auto
  then have "g x = g y"
    using `inj f`
    unfolding inj_def
    by auto
  then show "x = y"
    using `inj g`
    unfolding inj_def
    by auto
qed

lemma
  assumes "surj f" "surj g"
  shows "surj (f \<circ> g)"
  unfolding surj_def
proof
  fix y
  obtain z where "f z = y"
    using `surj f` unfolding surj_def
    by metis
  moreover
  obtain x where "g x = z"
    using `surj g` unfolding surj_def
    by metis
  ultimately
  show "\<exists> x. y = (f \<circ> g) x"
    unfolding comp_def
    by auto
qed

lemma
  assumes "inj f"
  shows "f ` (A \<inter> B) = f ` A \<inter> f ` B" (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof
    fix y
    assume "y \<in> f ` (A \<inter> B)"
    then obtain x where "x \<in> A \<inter> B" "f x = y"
      by auto
    then have "x \<in> A" "x \<in> B" "f x = y"
      by auto
    then have "y \<in> f ` A" "y \<in> f ` B"
      by auto
    then show "y \<in> f ` A \<inter> f ` B"
      by auto
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof
    fix y
    assume "y \<in> f ` A \<inter> f ` B"
    then have "y \<in> f ` A" "y \<in> f ` B"
      by auto
    then obtain x1 x2 where "y = f x1" "x1 \<in> A" "y = f x2" "x2 \<in> B"
      by blast
    then have "x1 = x2"
      using \<open>inj f\<close>
      unfolding inj_def
      by auto
    with \<open>y = f x1\<close> \<open>x1 \<in> A\<close> \<open>x2 \<in> B\<close>
    show "y \<in> f ` (A \<inter> B)"
      by auto
  qed
qed

lemma
  shows "A \<subseteq> f -` (f ` A)"
proof
  fix x
  assume "x \<in> A"
  then have "f x \<in> f ` A"
    by auto
  then show "x \<in> f -` (f ` A)"
    by auto
qed

lemma
  assumes "inj f"
  shows "f -` (f ` A) \<subseteq> A"
proof
  fix x
  assume "x \<in> f -` (f ` A)"
  then have "f x \<in> f ` A"
    by auto
  then obtain x' where "x' \<in> A" "f x' = f x"
    by auto
  then have "x' = x"
    using \<open>inj f\<close>
    unfolding inj_def
    by auto
  with \<open>x' \<in> A\<close> show "x \<in> A"
    by auto
qed

end