theory Numbers
  imports Main HOL.Real
begin

lemma 
  fixes x y :: real
  shows "(x + y)\<^sup>2 = x\<^sup>2 + 2*x*y + y\<^sup>2"
  by (simp add: power2_eq_square algebra_simps)

lemma
  fixes x y :: real
  shows "(x + y)\<^sup>2 = x\<^sup>2 + 2*x*y + y\<^sup>2" 
proof-
  have "(x + y)\<^sup>2 = (x + y)*(x + y)"
    by (subst power2_eq_square, rule refl)
  also have "... = x*(x + y) + y*(x + y)"
    by (subst distrib_right, rule refl)
  also have "... = (x*x + x*y) + y*(x + y)"
    by (subst distrib_left, rule refl)
  also have "... = (x*x + x*y) + (y*x + y*y)"
    by (subst distrib_left, rule refl)
  also have "... = ((x*x + x*y) + y*x) + y*y"
    by (subst add.assoc, rule refl)
  also have "... = (x*x + (x*y + y*x)) + y*y"
    by (subst add.assoc, rule refl)
  also have "... = (x*x + (x*y + x*y)) + y*y"
    by (subst mult.commute, rule refl)
  also have "... = x*x + 2*(x*y) + y*y"
    by simp
  also have "... = x*x + 2*x*y + y*y"
    by (subst mult.assoc, rule refl)
  also have "... = x\<^sup>2 + 2*x*y + y\<^sup>2"
    by ((subst power2_eq_square)+, rule refl)
  finally show ?thesis
    .
qed

thm nat.induct
thm less_induct

lemma
  fixes n::nat
  shows "(\<Sum> x \<in> {0..<n+1}. x) = n*(n+1) div 2"
proof (induction n)
  case 0
  show ?case
    by simp
next
  case (Suc n)
  have "(\<Sum> {0..<Suc n + 1}) = (\<Sum> {0..<n + 1} + (n + 1))"
    by simp
  also have "... = (n * (n + 1) div 2) + (n + 1)"
    using Suc
    by simp
  also have "... = Suc n * (Suc n + 1) div 2"
    by simp
  finally show ?case
    .
qed

lemma 
  fixes n :: nat
  shows "(3::nat) dvd 5^n + 2^(n+1)"
proof (induction n)
  case 0
  show ?case
    by (simp add: numeral_3_eq_3)
next
  case (Suc n)
  have "(5::nat) ^ Suc n + 2 ^ (Suc n + 1) = 5*(5^n) + 2*(2^Suc n)"
    by simp
  also have "... = (3*(5^n) + 2*(5^n)) + 2*(2^Suc n)"
    by simp
  also have "... = 3*(5^n) + 2*(5^n + 2^Suc n)"
    by simp
  finally 
  show ?case
    using Suc
    by (metis Suc_eq_plus1 dvd_add_times_triv_left_iff dvd_mult mult.commute)
qed


datatype nat = Zero ("\<zero>") | Suc nat

primrec add :: "nat \<Rightarrow> nat \<Rightarrow> nat" (infixl "\<oplus>" 100) where
  "x \<oplus> \<zero> = x"
| "x \<oplus> (Suc y) = Suc (x \<oplus> y)"

lemma add_zero_left [simp]:
  "\<zero> \<oplus> x = x"
  by (induction x) auto

lemma add_zero_left_readable_proof:
  "\<zero> \<oplus> x = x"
proof (induction x)
  case Zero
  then show ?case
    by (rule add.simps(1))
next
  case (Suc x)
  have "\<zero> \<oplus> (Suc x) = Suc (\<zero> \<oplus> x)"
    by (rule add.simps(2))
  also have "... = Suc x"
    by (subst Suc, rule refl)
  finally show ?case
    .
qed

lemma add_Suc_left [simp]:
  "Suc x \<oplus> y = Suc (x \<oplus> y)"
  by (induction y) auto

lemma add_assoc:
  "x \<oplus> (y \<oplus> z) = (x \<oplus> y) \<oplus> z"
  by (induction x) auto

lemma add_commute:
  "x \<oplus> y = y \<oplus> x"
  by (induction x) auto

primrec mult :: "nat \<Rightarrow> nat \<Rightarrow> nat" (infixl "\<otimes>" 101) where
  "x \<otimes> \<zero> = \<zero>"
| "x \<otimes> (Suc y) = x \<oplus> (x \<otimes> y)"

lemma mult_zero_left [simp]:
  "\<zero> \<otimes> x = \<zero>"
  by (induction x) auto

lemma mult_suc_left [simp]:
  "Suc x \<otimes> y = (x \<otimes> y) \<oplus> y"
  by (induction y) (auto simp add: add_assoc)

lemma distrib_left [simp]:
  "x \<otimes> (y \<oplus> z) = x \<otimes> y \<oplus> x \<otimes> z"
  by (induction x, simp, metis add_assoc add_commute mult_suc_left)

lemma distrib_left_readable_proof:
  "x \<otimes> (y \<oplus> z) = x \<otimes> y \<oplus> x \<otimes> z"
proof (induction x)
  case Zero
  show ?case 
    by ((subst mult_zero_left)+, subst add_zero_left, rule refl)
next
  case (Suc x)
  have "Suc x \<otimes> (y \<oplus> z) = x \<otimes> (y \<oplus> z) \<oplus> (y \<oplus> z)"
    by (subst mult_suc_left, rule refl)
  also have "... = x \<otimes> y \<oplus> x \<otimes> z \<oplus> (y \<oplus> z)"
    by (subst Suc, rule refl)
  also have "... = x \<otimes> y \<oplus> (x \<otimes> z \<oplus> (y \<oplus> z))"
    by (subst add_assoc, rule refl)
  also have "... = x \<otimes> y \<oplus> ((y \<oplus> z) \<oplus> x \<otimes> z)"
    by (subst add_commute, rule refl)
  also have "... = x \<otimes> y \<oplus> (y \<oplus> (z \<oplus> x \<otimes> z))"
    by (subst add_assoc, rule refl)
  also have "... = (x \<otimes> y \<oplus> y) \<oplus> (z \<oplus> x \<otimes> z)"
    by (subst add_assoc, rule refl)
  also have "... = (Suc x \<otimes> y) \<oplus> (z \<oplus> x \<otimes> z)"
    by (subst mult_suc_left, rule refl)
  also have "... = (Suc x \<otimes> y) \<oplus> (x \<otimes> z \<oplus> z)"
    by (subst add_commute, rule refl)
  also have "... = (Suc x \<otimes> y) \<oplus> (Suc x \<otimes> z)"
    by (subst mult_suc_left, rule refl)
  finally
  show ?case
    .
qed

lemma distrib_right [simp]:
  "(y \<oplus> z) \<otimes> x = y \<otimes> x \<oplus> z \<otimes> x"
  by (induction x, simp, metis add_assoc add_commute mult.simps(2))

lemma mult_assoc:
  "x \<otimes> (y \<otimes> z) = (x \<otimes> y) \<otimes> z"
  by (induction x) auto

lemma mult_commute:
  "x \<otimes> y = y \<otimes> x"
  by (induction x) (auto simp add: add_commute)


end