theory Textbook
  imports Main "HOL-Computational_Algebra.Primes" HOL.Real
begin

section \<open>Limits\<close>

definition is_limit :: "(real \<Rightarrow> real) \<Rightarrow> real \<Rightarrow> real \<Rightarrow> bool" where
  "is_limit f x0 a \<longleftrightarrow>
    (\<forall> \<epsilon> > 0. \<exists> \<delta> > 0. \<forall> x. \<bar>x - x0\<bar> < \<delta> \<longrightarrow> \<bar>f x - a\<bar> < \<epsilon>)"

lemma
  assumes "is_limit g x0 a" "is_limit h x0 b" 
          "\<forall> x. f x = g x + h x"
  shows "is_limit f x0 (a + b)"
  unfolding is_limit_def
proof safe
    fix \<epsilon> :: real
    assume "\<epsilon> > 0"
    then have "\<epsilon> / 2 > 0"
      by auto

    obtain \<delta>1 where  g: "\<delta>1 > 0" 
       "\<forall> x. \<bar>x - x0\<bar> < \<delta>1 \<longrightarrow> \<bar>g x - a\<bar> < \<epsilon>/2"
      using `is_limit g x0 a` `\<epsilon> / 2 > 0`
      unfolding is_limit_def
      by metis

    obtain \<delta>2 where h: "\<delta>2 > 0" "\<forall> x. \<bar>x - x0\<bar> < \<delta>2 \<longrightarrow> \<bar>h x - b\<bar> < \<epsilon>/2"
      using `is_limit h x0 b` `\<epsilon> / 2 > 0`
      unfolding is_limit_def
      by metis

    let ?\<delta> = "min \<delta>1 \<delta>2"
    have "\<forall>x. \<bar>x - x0\<bar> < ?\<delta> \<longrightarrow> \<bar>f x - (a + b)\<bar> < \<epsilon>"
    proof safe
      fix x
      assume "\<bar>x - x0\<bar> < ?\<delta>" 
      have "\<bar>f x - (a + b)\<bar> = \<bar>(g x - a) + (h x - b)\<bar>"
        using `\<forall> x. f x = g x + h x`
        by auto
      also have "... \<le> \<bar>g x - a\<bar> + \<bar>h x - b\<bar>"
        using abs_triangle_ineq by blast
      also have "... < \<epsilon> / 2 + \<epsilon> / 2"
        using g(2) h(2) \<open>\<bar>x - x0\<bar> < ?\<delta>\<close>                             
        by fastforce
      finally show "\<bar>f x - (a + b)\<bar> < \<epsilon>"
        by auto
    qed

    moreover

    have "?\<delta> > 0"
      using \<open>\<delta>1 > 0\<close> \<open>\<delta>2 > 0\<close>
      by linarith

    ultimately

    show "\<exists>\<delta>>0. \<forall>x. \<bar>x - x0\<bar> < \<delta> \<longrightarrow> \<bar>f x - (a + b)\<bar> < \<epsilon>"
      by blast
qed


section \<open>Infinitely many primes\<close>

term prime_factors

definition min_prime_factor :: "nat \<Rightarrow> nat" where
  "min_prime_factor n = Min (prime_factors n)"

lemma min_prime_factor_prime_factors:
  assumes "n > 1"
  shows "min_prime_factor n \<in> prime_factors n"
  by (metis Min_in One_nat_def Suc_le_eq assms finite_set_mset min_prime_factor_def nat_less_le prod_mset_empty prod_mset_prime_factorization_nat set_mset_eq_empty_iff)

lemma min_prime_factor_prime:
  assumes "n > 1"
  shows "prime (min_prime_factor n)"
  using assms min_prime_factor_prime_factors by blast

lemma min_prime_factor_dvd:
  assumes "n > 1"
  shows "min_prime_factor n dvd n"
  using assms min_prime_factor_prime_factors by blast

lemma min_prime_factor_gt_1:
  assumes "n > 1"
  shows "min_prime_factor n > 1"
  using assms min_prime_factor_prime prime_nat_iff' by blast

lemma infinitely_many_primes:
  fixes n :: nat
  shows "\<exists> p > n. prime p"
proof
  let ?N = "(fact n + 1)::nat"
  let ?p = "min_prime_factor ?N"

  show "?p > n \<and> prime ?p"
  proof
    show "?p > n"
    proof (rule ccontr)
      assume "\<not> ?thesis"
      then have "?p dvd fact n"
        using min_prime_factor_prime prime_dvd_fact_iff by auto
      moreover
      have "?p dvd ?N"
        using min_prime_factor_dvd by auto
      ultimately
      have "?p dvd 1"
        using dvd_add_right_iff by blast
      then show False
        by (metis Suc_eq_plus1 add.commute add_diff_cancel_left' cancel_comm_monoid_add_class.diff_cancel fact_nonzero less_one min_prime_factor_prime nat.simps(3) nat_neq_iff not_prime_unit)
    qed
  next
    show "prime ?p"
      by (simp add: min_prime_factor_prime)
  qed
qed

section \<open>Divisibility\<close>

lemma 
  fixes n :: nat
  shows "6 dvd (n^3 - n)"
proof-
  have "n^3 - n = (n - 1) * n * (n + 1)"
    by (simp add: diff_mult_distrib diff_mult_distrib2 power3_eq_cube)

  moreover

  have "3 dvd (n - 1) \<or> 3 dvd n \<or> 3 dvd (n + 1)"
  proof (cases "n mod 3 = 0")
    case True
    then show ?thesis
      by auto
  next
    case False
    show ?thesis
    proof (cases "n mod 3 = 1")
      case True
      then have "(n - 1) mod 3 = 0"
        by (metis diff_add_inverse2 div_mod_decomp dvd_imp_mod_0 dvd_triv_right)
      then show ?thesis
        by blast
    next
      case False
      with `n mod 3 \<noteq> 0`
      have "n mod 3 = 2"
        by (metis One_nat_def eval_nat_numeral(3) less_2_cases_iff less_Suc_eq mod_less_divisor)
      then have "(n + 1) mod 3 = 0"
        by (metis Suc_eq_plus1 eval_nat_numeral(3) mod_Suc)
      then show ?thesis
        by blast
    qed   
  qed
  then have "3 dvd (n - 1) * n * (n + 1)"
    by (metis dvd_mult dvd_mult2)

  moreover

  have "2 dvd n \<or> 2 dvd (n+1)"
  proof (cases "n mod 2 = 0")
    case True
    then show ?thesis
      by simp
  next
    case False
    then have "(n + 1) mod 2 = 0"
      by (meson dvd_imp_mod_0 even_plus_one_iff)
    then show ?thesis
      by simp
  qed
  then have "2 dvd ((n - 1) * n * (n + 1))"
    by auto

  moreover
  have "coprime (2::nat) 3"
    by auto

  ultimately

  show ?thesis
    using semiring_gcd_class.divides_mult[of 2 "n ^ 3 - n" 3]
    by simp
qed

end